//
// 	NetFilterSDK 
// 	Copyright (C) Vitaly Sidorov
//	All rights reserved.
//
//	This file is a part of the NetFilter SDK.
//	The code and information is provided "as-is" without
//	warranty of any kind, either expressed or implied.
//


#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/un.h>
#include <sys/errno.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/time.h> 
#include <sys/errno.h>
#include <sys/resource.h> 
#include <sys/sysinfo.h>

#include "nfproxy.h"

NF_STATUS nf_addRouteRules();
NF_STATUS nf_deleteRouteRules();

#if defined(_DEBUG) || defined(_RELEASE_LOG)
DBGLogger_ DBGLogger_::dbgLogProxy;
#endif

AutoCriticalSection g_cs;
NFProxy_ptr			g_proxy;
std::string			g_driverName;
int					g_nThreads = 1;

static bool startProxy(NF_EventHandler * pHandler)
{
	try
	{
		DbgPrint("create proxy");
		g_proxy = NFProxy_ptr(new NFProxy(pHandler));

		DbgPrint("start proxy");
		g_proxy->start(g_nThreads);
		return true;
	}
	catch (std::exception& e)
	{
		DbgPrint("IOThread::start() Exception: %s", e.what());
	}

	g_proxy.reset();

	return false;
}

static void stopProxy()
{
	if (g_proxy)
	{
		g_proxy->stop();
		g_proxy.reset();
	}
}

NFAPI_API NF_STATUS NFAPI_NS nf_init(const char * driverName, NF_EventHandler * pHandler)
{
	NF_STATUS status = NF_STATUS_SUCCESS;
    
#if defined(_DEBUG) || defined(_RELEASE_LOG)
	if (!DBGLogger_::instance().init("NFProxyLog.txt"))
	{
		printf("Unable to start the debug log.\n");
	}
#endif

	AutoLock lock(g_cs);

	DbgPrint("nf_init");

	g_driverName = driverName;

	if (!startProxy(pHandler))
	{
		DbgPrint("startProxy failed");
		status = NF_STATUS_FAIL;
	}

	if (status == NF_STATUS_SUCCESS)
	{
		nf_addRouteRules();
	}

	return status;
}

NFAPI_API void NFAPI_NS nf_free()
{
	DbgPrint("nf_free");

	try {
		stopProxy();
	}
	catch (std::exception& e)
	{
		DbgPrint("nf_free() Exception: %s", e.what());
	}

	nf_deleteRules();
	nf_deleteRouteRules();

	DbgPrint("nf_free stopped");

#if defined(_DEBUG) || defined(_RELEASE_LOG)
	DBGLogger_::instance().free();
#endif
	return;
}

NFAPI_API void NFAPI_NS 
nf_setOptions(int nThreads, int flags)
{
	g_nThreads = nThreads;
	
	if (g_nThreads <= 0)
	{
		g_nThreads = get_nprocs();
		if (g_nThreads <= 0)
			g_nThreads = 1;
	} else
	if (g_nThreads > 1024)
	{
		g_nThreads = 1024;
	}
}

NFAPI_API NF_STATUS NFAPI_NS nf_tcpPostSend(ENDPOINT_ID id, const char * buf, int len)
{
	return g_proxy->postSend(id, buf, len)? NF_STATUS_SUCCESS : NF_STATUS_FAIL;
}

NFAPI_API NF_STATUS NFAPI_NS nf_tcpPostReceive(ENDPOINT_ID id, const char * buf, int len)
{
	return g_proxy->postReceive(id, buf, len)? NF_STATUS_SUCCESS : NF_STATUS_FAIL;
}

NFAPI_API NF_STATUS NFAPI_NS nf_tcpClose(ENDPOINT_ID id)
{
	return g_proxy->deleteSession(id)? NF_STATUS_SUCCESS : NF_STATUS_FAIL;
}

NFAPI_API NF_STATUS NFAPI_NS nf_tcpSetConnectionState(ENDPOINT_ID id, int suspended)
{
	return g_proxy->setState(id, suspended? true : false)? NF_STATUS_SUCCESS : NF_STATUS_FAIL;
}

NFAPI_API NF_STATUS NFAPI_NS nf_udpPostSend(ENDPOINT_ID id, const unsigned char * remoteAddress, const char * buf, int len, PNF_UDP_OPTIONS options)
{
	return NF_STATUS_SUCCESS;
}

NFAPI_API NF_STATUS NFAPI_NS nf_udpPostReceive(ENDPOINT_ID id, const unsigned char * remoteAddress, const char * buf, int len, PNF_UDP_OPTIONS options)
{
	return NF_STATUS_SUCCESS;
}

NFAPI_API NF_STATUS NFAPI_NS nf_udpSetConnectionState(ENDPOINT_ID id, int suspended)
{
	return NF_STATUS_SUCCESS;
}

NFAPI_API unsigned long NFAPI_NS 
nf_setTCPTimeout(unsigned long timeout)
{
	g_proxy->setTimeout(timeout);
	return 0;
}

/**
 *	Disables indicating TCP packets to user mode for the specified endpoint
 *  @param id Socket identifier
 */
NFAPI_API NF_STATUS NFAPI_NS 
nf_tcpDisableFiltering(ENDPOINT_ID id)
{
	return g_proxy->disableEvents(id, true)? NF_STATUS_SUCCESS : NF_STATUS_FAIL;
}

/**
* Returns in pConnInfo the properties of TCP connection with specified id.
**/
NFAPI_API NF_STATUS NFAPI_NS
nf_getTCPConnInfo(ENDPOINT_ID id, PNF_TCP_CONN_INFO pConnInfo)
{
	return g_proxy->getConnInfo(id, pConnInfo)? NF_STATUS_SUCCESS : NF_STATUS_FAIL;
}


NFAPI_API unsigned long NFAPI_NS 
nf_getConnCount()
{
	return g_proxy->getConnCount();
}

NFAPI_API unsigned short NFAPI_NS 
nf_getProxyPort()
{
	return g_proxy->getPort();
}

NFAPI_API NF_STATUS NFAPI_NS 
nf_getUid(ENDPOINT_ID id, unsigned int * pUid)
{
	return g_proxy->getUid(id, pUid)? NF_STATUS_SUCCESS : NF_STATUS_FAIL;
}

NFAPI_API NF_STATUS NFAPI_NS 
nf_getUserName(unsigned int uid, char * buf, int len)
{
	struct passwd * pwd;
	NF_STATUS res = NF_STATUS_FAIL;

	pwd = getpwuid(uid);
	if (pwd)
	{
		snprintf(buf, len, "%s", pwd->pw_name);
		res = NF_STATUS_SUCCESS;
	}

	return res;
}

NFAPI_API NF_STATUS NFAPI_NS 
nf_getProcessName(unsigned int pid, char * buf, int len)
{
    char path[PATH_MAX];
	FILE * f;
	NF_STATUS res = NF_STATUS_FAIL;

	snprintf(path, sizeof(path), "/proc/%u/cmdline", pid);

	f = fopen(path, "r");
	if (!f)
		return res;

    if (fgets(buf, len, f))
		res = NF_STATUS_SUCCESS;

	fclose(f);

	return res;
}

NFAPI_API NF_STATUS NFAPI_NS 
nf_requireFileLimit(int file_limit)
{
	struct rlimit lim;

	if (getrlimit(RLIMIT_NOFILE, &lim) != 0)
	{
		return NF_STATUS_FAIL;
	}

	if (lim.rlim_cur >= file_limit)
		return NF_STATUS_SUCCESS;

	lim.rlim_cur = lim.rlim_max = file_limit;

	if (setrlimit(RLIMIT_NOFILE, &lim) != 0)
	{
		return NF_STATUS_FAIL;
	}

	if (getrlimit(RLIMIT_NOFILE, &lim) != 0)
	{
		return NF_STATUS_FAIL;
	}

	return NF_STATUS_SUCCESS;
}
