//
// 	NetFilterSDK 
// 	Copyright (C) Vitaly Sidorov
//	All rights reserved.
//
//	This file is a part of the NetFilter SDK.
//	The code and information is provided "as-is" without
//	warranty of any kind, either expressed or implied.
//

#ifndef _NFPROXY_H
#define _NFPROXY_H

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <sys/types.h>
#include <ifaddrs.h>

#include <boost/shared_ptr.hpp>
#include <boost/enable_shared_from_this.hpp>
#include <boost/thread/recursive_mutex.hpp>
#include <boost/thread/thread.hpp>
#include <boost/bind.hpp>
#include <boost/asio.hpp>

#include <queue>
#include <map>

#include "nfapi_linux.h"
#include "sync.h"

#include <linux/netfilter_ipv4.h>

#define MARK_OUT 0x1006
#define MARK_IN_S "0x800"
#define MARK_IN_S_ "0x806"
#define MARK_PASS_S "0x1400"
#define MARK_PASS 0x1400
#define MARK_TABLE "800"

#define PACKET_LEN	8192

using namespace nfapi;

#include "dbglogger.h"

using boost::asio::ip::tcp;

class NFSessionContainer
{
public:
	virtual bool deleteSession(ENDPOINT_ID id) = 0;
	virtual bool deleteSession_unsafe(ENDPOINT_ID id) = 0;
	virtual unsigned short getPort() = 0;
	virtual unsigned short getPortIPv6() = 0;
};

class NFProxy;

class NFSession :
    public boost::enable_shared_from_this<NFSession>
{
public:
  NFSession(NFSessionContainer & proxy, 
	  NF_EventHandler * pEventHandler,
	  boost::asio::io_context & io_context, 
	  ENDPOINT_ID id,
	  unsigned int timeout)
    : m_pEventHandler(pEventHandler),
    m_socketLocal(io_context), 
	m_socketRemote(io_context), 
	m_timer(io_context),
	m_proxy(proxy), 
	m_id(id),
	m_suspend(false),
	m_wantSuspend(false),
	m_recvLocalInProgress(false),
	m_recvRemoteInProgress(false),
	m_sendLocalInProgress(false),
	m_sendRemoteInProgress(false),
	m_remoteState(SS_NONE),
	m_localState(SS_NONE),
	m_firstPacket(true),
	m_timeout(timeout),
	m_closed(false),
	m_eventsDisabled(false),
	m_uid(0)
  {
  }

	~NFSession()
	{
	}

  tcp::socket& getLocalSocket()
  {
		return m_socketLocal;
  }

  ENDPOINT_ID getId()
  {
	  return m_id;
  }

	NF_TCP_CONN_INFO & getConnInfo()
	{
		return m_ci;
	}

	void close(const boost::system::error_code& error)
	{
		if (!error)
		{
			DbgPrint("NFSession::close[%llu]", m_id);
			m_socketLocal.close();
			m_socketRemote.close();
		}
	}

	void setTimeout()
	{
		if (m_closed)
			return;

		if (m_timeout > 0)
		{
			m_timer.expires_from_now(boost::posix_time::seconds(m_timeout));

			m_timer.async_wait(boost::bind(&NFSession::close, 
				shared_from_this(),
				boost::asio::placeholders::error));
		}
	}

	bool start()
	{
		DbgPrint("NFSession::start()");

		boost::asio::socket_base::keep_alive option(true);
		m_socketLocal.set_option(option);

		boost::asio::ip::tcp::no_delay nd_option(true);
		m_socketLocal.set_option(nd_option);

		boost::asio::ip::tcp::endpoint remoteEndpoint = m_socketLocal.remote_endpoint();
		boost::asio::ip::tcp::endpoint localEndpoint = m_socketLocal.local_endpoint();

		m_ci.processId = 0;
		m_ci.direction = NF_D_OUT;
		m_ci.filteringFlag = NF_FILTER;
		m_ci.ip_family = remoteEndpoint.protocol().family();

		boost::asio::ip::tcp::endpoint real_remoteEndpoint;

		if (remoteEndpoint.protocol().family() == AF_INET)
		{
			memcpy(&m_ci.localAddress, remoteEndpoint.data(), sizeof(sockaddr_in));
			memcpy(&m_ci.remoteAddress, localEndpoint.data(), sizeof(sockaddr_in));
			
#ifdef _DEBUG
			{
				std::string sLocalAddr = inet_ntoa(((sockaddr_in*)m_ci.localAddress)->sin_addr);
				std::string sRemoteAddr = inet_ntoa(((sockaddr_in*)m_ci.remoteAddress)->sin_addr);

				DbgPrint("NFSession::start() connect local=%s:%d remote=%s:%d", 
								sLocalAddr.c_str(), ntohs(((sockaddr_in*)m_ci.localAddress)->sin_port),
								sRemoteAddr.c_str(), ntohs(((sockaddr_in*)m_ci.remoteAddress)->sin_port));
			}
#endif
			if (isLocalAddress((sockaddr*)localEndpoint.data()))
			{
				unsigned short remotePort = ntohs(((sockaddr_in*)m_ci.remoteAddress)->sin_port);

				if (m_proxy.getPort() == remotePort || m_proxy.getPortIPv6() == remotePort)
				{
					DbgPrint("NFSession::start() inbound connection to local proxy address, closing");
					return false;
				}

				if ((((sockaddr_in*)localEndpoint.data())->sin_addr.s_addr !=
					((sockaddr_in*)remoteEndpoint.data())->sin_addr.s_addr))
				{
					DbgPrint("NFSession::start() inbound connection");
					m_ci.direction = NF_D_IN;
				}
			}

			TCP_PROC_INFO tpi;

			if (m_ci.direction == NF_D_OUT)
			{
				if (getTCPProcInfo(AF_INET, ntohs(((sockaddr_in*)m_ci.localAddress)->sin_port), &tpi))
				{
					m_ci.processId = tpi.pid;
					m_uid = tpi.uid;
				}

				m_pEventHandler->tcpConnectRequest(m_id, &m_ci);
		
				if (m_ci.filteringFlag & NF_BLOCK)
					return false;

				real_remoteEndpoint = boost::asio::ip::tcp::endpoint(
						boost::asio::ip::address_v4(htonl(((sockaddr_in*)m_ci.remoteAddress)->sin_addr.s_addr)), 
						htons(((sockaddr_in*)m_ci.remoteAddress)->sin_port));
			} else
			{
				real_remoteEndpoint = boost::asio::ip::tcp::endpoint(
						boost::asio::ip::address_v4(htonl(((sockaddr_in*)m_ci.remoteAddress)->sin_addr.s_addr)), 
						htons(((sockaddr_in*)m_ci.remoteAddress)->sin_port));

				memcpy(m_ci.localAddress, m_ci.remoteAddress, sizeof(m_ci.localAddress));
				memcpy(m_ci.remoteAddress, remoteEndpoint.data(), sizeof(sockaddr_in));

				if (getTCPProcInfo(AF_INET, ntohs(((sockaddr_in*)m_ci.localAddress)->sin_port), &tpi))
				{
					m_ci.processId = tpi.pid;
					m_uid = tpi.uid;
				}

				if (m_ci.processId == getpid())
					return false;
			}
		} else
		{
			memcpy(&m_ci.localAddress, remoteEndpoint.data(), sizeof(sockaddr_in6));
			memcpy(&m_ci.remoteAddress, localEndpoint.data(), sizeof(sockaddr_in6));

#ifdef _DEBUG            
			{
				char sAddr[100];
				std::string sLocalAddr;
				std::string sRemoteAddr;

				if (inet_ntop(AF_INET6, &((sockaddr_in6*)m_ci.localAddress)->sin6_addr, sAddr, sizeof(sAddr)))
				{
					sLocalAddr = sAddr;
				}
				if (inet_ntop(AF_INET6, &((sockaddr_in6*)m_ci.remoteAddress)->sin6_addr, sAddr, sizeof(sAddr)))
				{
					sRemoteAddr = sAddr;
				}

				DbgPrint("NFSession::start() connect local=[%s]:%d remote=[%s]:%d", 
								sLocalAddr.c_str(), ntohs(((sockaddr_in6*)m_ci.localAddress)->sin6_port),
								sRemoteAddr.c_str(), ntohs(((sockaddr_in6*)m_ci.remoteAddress)->sin6_port));
			}
#endif

			if (isLocalAddress((sockaddr*)localEndpoint.data()))
			{
				unsigned short remotePort = ntohs(((sockaddr_in6*)m_ci.remoteAddress)->sin6_port);

				if (m_proxy.getPort() == remotePort || m_proxy.getPortIPv6() == remotePort)
				{
					DbgPrint("NFSession::start() inbound connection to local proxy address, closing");
					return false;
				}

				if (memcmp(&((sockaddr_in6*)localEndpoint.data())->sin6_addr,
						&((sockaddr_in6*)remoteEndpoint.data())->sin6_addr,
						16) != 0)
				{
					DbgPrint("NFSession::start() inbound connection");
					m_ci.direction = NF_D_IN;
				}
			}

			TCP_PROC_INFO tpi;

			if (m_ci.direction == NF_D_OUT)
			{
				if (getTCPProcInfo(AF_INET6, ntohs(((sockaddr_in6*)m_ci.localAddress)->sin6_port), &tpi))
				{
					m_ci.processId = tpi.pid;
					m_uid = tpi.uid;
				}

				m_pEventHandler->tcpConnectRequest(m_id, &m_ci);
		
				boost::asio::ip::address_v6::bytes_type ipv6addr;
				unsigned char *raddr = reinterpret_cast<unsigned char*>(&((sockaddr_in6*)m_ci.remoteAddress)->sin6_addr);
				std::copy(raddr, raddr + 16, ipv6addr.data());

				real_remoteEndpoint = boost::asio::ip::tcp::endpoint(
							boost::asio::ip::address_v6(ipv6addr), 
							htons(((sockaddr_in6*)m_ci.remoteAddress)->sin6_port));
			} else
			{
				boost::asio::ip::address_v6::bytes_type ipv6addr;
				unsigned char *raddr = reinterpret_cast<unsigned char*>(&((sockaddr_in6*)m_ci.remoteAddress)->sin6_addr);
				std::copy(raddr, raddr + 16, ipv6addr.data());

				real_remoteEndpoint = boost::asio::ip::tcp::endpoint(
							boost::asio::ip::address_v6(ipv6addr), 
							htons(((sockaddr_in6*)m_ci.remoteAddress)->sin6_port));

				memcpy(m_ci.localAddress, m_ci.remoteAddress, sizeof(m_ci.localAddress));
				memcpy(m_ci.remoteAddress, remoteEndpoint.data(), sizeof(sockaddr_in6));

				if (getTCPProcInfo(AF_INET6, ntohs(((sockaddr_in6*)m_ci.localAddress)->sin6_port), &tpi))
				{
					m_ci.processId = tpi.pid;
					m_uid = tpi.uid;
				}

				if (m_ci.processId == getpid())
					return false;
			}
		}

		m_recvLocalInProgress = false;

		int remoteSocket = socket(remoteEndpoint.protocol().family(), SOCK_STREAM | SOCK_NONBLOCK, 0);
		
		m_socketRemote.assign(remoteEndpoint.protocol().family() == AF_INET? boost::asio::ip::tcp::v4() : boost::asio::ip::tcp::v6(), remoteSocket);

		setSocketOptions(m_socketRemote.native_handle());

		m_socketRemote.async_connect(real_remoteEndpoint, 
							boost::bind(&NFSession::connect_complete, 
								shared_from_this(), 
								boost::asio::placeholders::error));

		setTimeout();

		return true;
	}

	bool postSend(const char * buf, size_t len)
	{
		if (m_ci.direction == NF_D_OUT)
		{
			return _postSend(buf, len);
		} else
		{
			return _postReceive(buf, len);
		}
	}

	bool postReceive(const char * buf, size_t len)
	{
		if (m_ci.direction == NF_D_OUT)
		{
			return _postReceive(buf, len);
		} else
		{
			return _postSend(buf, len);
		}
	}

	void setState(bool suspend)
	{
		AutoLock lock(m_cs);

		if (m_wantSuspend == suspend)
			return;

		m_wantSuspend = suspend;

		setStateInternal(suspend);
	}

	void setStateInternal(bool suspend)
	{
		if (m_closed)
			return;

		if (!suspend)
		{
			if (!m_recvLocalInProgress && (m_localState == SS_NONE))
			{
				m_recvLocalInProgress = true;

				m_socketLocal.async_read_some(boost::asio::buffer(m_recvBufLocal, sizeof(m_recvBufLocal)),
					boost::bind(&NFSession::recv_local_complete, shared_from_this(),
						boost::asio::placeholders::error,
						boost::asio::placeholders::bytes_transferred));

			}
		
			if (!m_recvRemoteInProgress && (m_remoteState == SS_NONE))
			{
				m_recvRemoteInProgress = true;

				m_socketRemote.async_read_some(boost::asio::buffer(m_recvBufRemote, sizeof(m_recvBufRemote)),
					boost::bind(&NFSession::recv_remote_complete, shared_from_this(),
						boost::asio::placeholders::error,
						boost::asio::placeholders::bytes_transferred));

			}

		}

		m_suspend = suspend;
	}

	void disableEvents(bool disable)
	{
		AutoLock lock(m_cs);
		m_eventsDisabled = disable;
	}

	unsigned int getUid()
	{
		AutoLock lock(m_cs);
		return m_uid;	
	}

private:

	typedef struct _TCP_PROC_INFO
	{
		pid_t pid;
		uid_t uid;
		ino_t iNode;	
	} TCP_PROC_INFO;

	bool findProcessId(TCP_PROC_INFO * pInfo)
	{
		pid_t pid = 0;
		char path[PATH_MAX];
		int pathLen;
		char link_src[PATH_MAX];
		struct stat statbuf;
		char* endptr;
		bool pidFound = false;

		DbgPrint("NFSession::findProcessId[%llu] inode=%d uid=%d", m_id, pInfo->iNode, pInfo->uid);

		DIR *dir = opendir("/proc");
		if (dir == NULL) 
		{
			return false;
		}
	    
		struct dirent* de;
		while ((de = readdir(dir))) 
		{
			if (!strcmp(de->d_name, ".") || !strcmp(de->d_name, ".."))
				 continue;

			pid = strtol(de->d_name, &endptr, 10);

			if (*endptr != '\0')
				continue;

			snprintf(path, sizeof(path), "/proc/%d/fd", pid);

    		DIR *dirFd = opendir(path);
			if (dirFd == NULL) 
			{
				DbgPrint("opendir failed for %s", path);
    			continue;
			}

			strncpy(link_src, path, sizeof(link_src));
			strncat(link_src, "/", sizeof(link_src));
			pathLen = strlen(path)+1;

			struct dirent* deFd;
			while ((deFd = readdir(dirFd))) 
			{
    			if (!strcmp(deFd->d_name, ".") || !strcmp(deFd->d_name, ".."))
        			continue;

				link_src[pathLen] = '\0';

				strncat(link_src, deFd->d_name, sizeof(link_src));

				if (stat(link_src, &statbuf) == 0)
				{
					if (S_ISSOCK(statbuf.st_mode))
					{
						if (pInfo->iNode == statbuf.st_ino)
						{
							pInfo->pid = pid;
							pidFound = true;
							break;
						}
					}
				}
			}

    		closedir(dirFd);

			if (pidFound)
				break;
		}

		closedir(dir);

		return pidFound;
	}


	bool getTCPProcInfo(int ip_family, unsigned short localPortToFind, TCP_PROC_INFO * pInfo)
	{
		FILE * f;
		unsigned long rxq, txq, time_len, retr, inode;
		int num, local_port, rem_port, d, state, uid, timer_run, timeout;
		char rem_addr[128], local_addr[128];
		char buffer[8192];	
		bool res = false;

		DbgPrint("NFSession::getTCPProcInfo[%llu] port %x", m_id, localPortToFind);

		if (ip_family == AF_INET)
		{
			f = fopen("/proc/net/tcp", "r");
		} else
		{
			f = fopen("/proc/net/tcp6", "r");
		}
		if (!f)
		{
			DbgPrint("NFSession::getTCPProcInfo[%llu] unable to open /proc/net/tcp[6]", m_id);
			return false;
		}

		fgets(buffer, sizeof(buffer), f);
		
		do {						
			if (!fgets(buffer, sizeof(buffer), f))
    			break;

			num = sscanf(buffer,
    			"%d: %64[0-9A-Fa-f]:%X %64[0-9A-Fa-f]:%X %X %lX:%lX %X:%lX %lX %d %d %lu %*s\n",
				 &d, local_addr, &local_port, rem_addr, &rem_port, &state,
				 &txq, &rxq, &timer_run, &time_len, &retr, &uid, &timeout, &inode);

    		if (num < 11) 
			{
				DbgPrint("NFSession::getTCPProcInfo[%llu] sscanf failed %s", m_id, buffer);
				break;
			}

			if (localPortToFind == local_port)
			{
				pInfo->iNode = inode;
				pInfo->uid = uid;

				if (findProcessId(pInfo))
				{
					res = true;
				}
				break;
			}

		} while (!feof(f));		


		fclose(f);

		return res;
	}

	bool _postSend(const char * buf, size_t len)
	{
		DbgPrint("NFSession::postSend[%llu] len=%d", m_id, len);

		AutoLock lock(m_cs);

		if (m_closed)
			return true;

		if (len == 0)
		{
			if (m_localState != SS_DISCONNECTED)
				m_localState = SS_MUST_CLOSE;

			if (m_sendBufRemote.empty())
			{
				DbgPrint("NFSession::postSend[%llu] disconnecting", m_id);

				try {
					boost::system::error_code ec;
					m_socketRemote.shutdown(boost::asio::socket_base::shutdown_send, ec);
					if (ec)
					{
						m_remoteState = SS_DISCONNECTED;
					}
				} catch (std::exception &)
				{
					m_remoteState = SS_DISCONNECTED;
				}

				m_localState = SS_DISCONNECTED;

				DbgPrint("NFSession::postSend[%llu] localState=%d remoteState=%d", m_id, m_localState, m_remoteState);

				if ((m_localState == SS_DISCONNECTED) && (m_remoteState == SS_DISCONNECTED))
				{
					DbgPrint("NFSession::postSend[%llu] disconnected", m_id);
					m_timer.cancel();
					m_closed = true;


					AutoUnlock lock(m_cs);
					m_proxy.deleteSession(m_id);
				}
			}

			return true;
		}

		setTimeout();

		m_sendBufRemote.push(tBuffer_ptr(new tBuffer(buf, buf+len)));
		
		if (((m_sendBufLocal.size() > 4) || (m_sendBufRemote.size() > 4)) &&
			(m_localState == SS_NONE && m_remoteState == SS_NONE))
		{
			setStateInternal(true);
		} else
		{
			setStateInternal(m_wantSuspend);
		}

		if (!m_sendRemoteInProgress)
		{
			m_sendRemoteInProgress = true;

			boost::asio::async_write(m_socketRemote,
				  boost::asio::buffer(&(*m_sendBufRemote.front())[0], m_sendBufRemote.front()->size()),
				  boost::bind(&NFSession::send_remote_complete, 
					shared_from_this(),
					boost::asio::placeholders::error,
					boost::asio::placeholders::bytes_transferred));

		}

		return true;
	}

	bool _postReceive(const char * buf, size_t len)
	{
		DbgPrint("NFSession::postReceive[%llu] len=%d", m_id, len);

		AutoLock lock(m_cs);

		if (m_closed)
			return true;

		if (len == 0)
		{
			if (m_remoteState != SS_DISCONNECTED)
				m_remoteState = SS_MUST_CLOSE;

			if (m_sendBufLocal.empty())
			{
				DbgPrint("NFSession::postReceive[%llu] disconnecting", m_id);

				try {
					m_socketLocal.shutdown(boost::asio::socket_base::shutdown_send);
				} catch (std::exception &)
				{
				}

				m_remoteState = SS_DISCONNECTED;

				DbgPrint("NFSession::postReceive[%llu] localState=%d remoteState=%d", m_id, m_localState, m_remoteState);

				if ((m_localState == SS_DISCONNECTED) && (m_remoteState == SS_DISCONNECTED))
				{
					DbgPrint("NFSession::postReceive[%llu] disconnected", m_id);
					m_timer.cancel();
					m_closed = true;

					AutoUnlock lock(m_cs);
					m_proxy.deleteSession(m_id);
				}

			}

			return true;
		}

		setTimeout();

		m_sendBufLocal.push(tBuffer_ptr(new tBuffer(buf, buf+len)));
		
		if (((m_sendBufLocal.size() > 4) || (m_sendBufRemote.size() > 4)) &&
			(m_localState == SS_NONE && m_remoteState == SS_NONE))
		{
			setStateInternal(true);
		} else
		{
			setStateInternal(m_wantSuspend);
		}

		if (!m_sendLocalInProgress)
		{
			m_sendLocalInProgress = true;

			boost::asio::async_write(m_socketLocal,
				  boost::asio::buffer(&(*m_sendBufLocal.front())[0], m_sendBufLocal.front()->size()),
				  boost::bind(&NFSession::send_local_complete, 
					shared_from_this(),
					boost::asio::placeholders::error,
					boost::asio::placeholders::bytes_transferred));

		}

		return true;
	}
	
	void connect_complete(const boost::system::error_code& error)
	{
		DbgPrint("NFSession::connect_complete[%llu]", m_id);

		if (!error)
		{
			setTimeout();

			m_pEventHandler->tcpConnected(m_id, &m_ci);

//			boost::asio::socket_base::keep_alive option(true);
//			m_socketRemote.set_option(option);

			boost::asio::ip::tcp::no_delay nd_option(true);
			m_socketRemote.set_option(nd_option);

			if (!m_recvLocalInProgress)
			{
				m_recvLocalInProgress = true;
				m_socketLocal.async_read_some(boost::asio::buffer(m_recvBufLocal, sizeof(m_recvBufLocal)),
					boost::bind(&NFSession::recv_local_complete, shared_from_this(),
						boost::asio::placeholders::error,
						boost::asio::placeholders::bytes_transferred));
			}

			if (!m_recvRemoteInProgress)
			{
				m_recvRemoteInProgress = true;
				m_socketRemote.async_read_some(boost::asio::buffer(m_recvBufRemote, sizeof(m_recvBufRemote)),
					boost::bind(&NFSession::recv_remote_complete, shared_from_this(),
						boost::asio::placeholders::error,
						boost::asio::placeholders::bytes_transferred));
			}
		} else
		{
			DbgPrint("NFSession::connect_complete[%llu] fail", m_id);
			m_timer.cancel();
			m_closed = true;
			m_proxy.deleteSession(m_id);
		}
	}
  
	void recv_local_complete(const boost::system::error_code& error, size_t bytes_transferred)
	{
		AutoLock lock(m_cs);

		if (m_closed)
			return;

		if (!error)
		{
			DbgPrint("NFSession::recv_local_complete[%llu] %d", m_id, bytes_transferred);

			m_firstPacket = false;

			setTimeout();

			{
				AutoUnlock lock(m_cs);
				if (m_eventsDisabled)
				{
					_postSend(m_recvBufLocal, bytes_transferred);
				} else
				{
					if (m_ci.direction == NF_D_OUT)
					{
						m_pEventHandler->tcpSend(m_id, m_recvBufLocal, bytes_transferred);
					} else
					{
						m_pEventHandler->tcpReceive(m_id, m_recvBufLocal, bytes_transferred);
					}
				}
			}

			if (m_suspend)
			{
				m_recvLocalInProgress = false;
			} else
			{
				if (bytes_transferred > 0)
				{
					m_recvLocalInProgress = true;
	
					m_socketLocal.async_read_some(boost::asio::buffer(m_recvBufLocal, sizeof(m_recvBufLocal)),
						boost::bind(&NFSession::recv_local_complete, shared_from_this(),
							boost::asio::placeholders::error,
							boost::asio::placeholders::bytes_transferred));
				}
			}
		} else
		{
			DbgPrint("NFSession::recv_local_complete[%llu] disconnected, err=%d", m_id, error);

			if (m_firstPacket)
			{
				m_timer.cancel();
				m_closed = true;

				AutoUnlock lock(m_cs);
				m_proxy.deleteSession(m_id);
				DbgPrint("NFSession::recv_local_complete[%llu] disconnected and closed", m_id);
			} else
			{
				if (m_localState == SS_NONE)
				{
                    m_localState = SS_DISCONNECTING;
					
					AutoUnlock lock(m_cs);

					if (m_eventsDisabled)
					{
						_postSend(NULL, 0);
					} else
					{
						if (m_ci.direction == NF_D_OUT)
						{
							m_pEventHandler->tcpSend(m_id, NULL, 0);
						} else
						{
							m_pEventHandler->tcpReceive(m_id, NULL, 0);
						}
					}
				}

				setStateInternal(false);
		
				m_recvLocalInProgress = false;
			}
		}
	}

	void recv_remote_complete(const boost::system::error_code& error, size_t bytes_transferred)
	{
		AutoLock lock(m_cs);

		if (m_closed)
			return;

		if (!error)
		{
			DbgPrint("NFSession::recv_remote_complete[%llu] %d", m_id, bytes_transferred);

			setTimeout();

			{
				AutoUnlock lock(m_cs);
				if (m_eventsDisabled)
				{
					_postReceive(m_recvBufRemote, bytes_transferred);
				} else
				{
					if (m_ci.direction == NF_D_OUT)
					{
						m_pEventHandler->tcpReceive(m_id, m_recvBufRemote, bytes_transferred);
					} else
					{
						m_pEventHandler->tcpSend(m_id, m_recvBufRemote, bytes_transferred);
					}
				}
			}

			if (m_suspend)
			{
				m_recvRemoteInProgress = false;
			} else
			{
				if (bytes_transferred > 0)
				{
					m_recvRemoteInProgress = true;

					m_socketRemote.async_read_some(boost::asio::buffer(m_recvBufRemote, sizeof(m_recvBufRemote)),
						boost::bind(&NFSession::recv_remote_complete, shared_from_this(),
							boost::asio::placeholders::error,
							boost::asio::placeholders::bytes_transferred));
				}
			}
		} else
		{
			DbgPrint("NFSession::recv_remote_complete[%llu] disconnected, err=%d", m_id, error.value());

			m_recvRemoteInProgress = false;
	
			if (m_remoteState == SS_NONE)
			{
				m_remoteState = SS_DISCONNECTING;

				AutoUnlock lock(m_cs);
				if (m_eventsDisabled)
				{
					_postReceive(NULL, 0);
				} else
				{
					if (m_ci.direction == NF_D_OUT)
					{
						m_pEventHandler->tcpReceive(m_id, NULL, 0);
					} else
					{
						m_pEventHandler->tcpSend(m_id, NULL, 0);
					}
				}
			}

			setStateInternal(false);
		}
	}

	void send_local_complete(const boost::system::error_code& error, size_t bytes_transferred)
	{
		AutoLock lock(m_cs);

		if (!error)
		{
			DbgPrint("NFSession::send_local_complete[%llu]", m_id);

			m_sendLocalInProgress = false;

			setTimeout();

			if (!m_sendBufLocal.empty())
				m_sendBufLocal.pop();

			if (((m_sendBufLocal.size() > 4) || (m_sendBufRemote.size() > 4)) &&
				(m_localState == SS_NONE && m_remoteState == SS_NONE))
			{
			} else
			{
				setStateInternal(m_wantSuspend);
			}

			if (m_sendBufLocal.empty())
			{
				if (!m_eventsDisabled)
				{
					AutoUnlock lock(m_cs);
	
					if (m_ci.direction == NF_D_OUT)
						m_pEventHandler->tcpCanReceive(m_id);
					else
						m_pEventHandler->tcpCanSend(m_id);
				}

				if (m_sendBufLocal.empty() && (m_remoteState == SS_MUST_CLOSE))
				{
					DbgPrint("NFSession::send_local_complete[%llu] disconnecting", m_id);

					try {
						m_socketLocal.shutdown(boost::asio::socket_base::shutdown_send);
					} catch (std::exception &)
					{
    				}

					m_remoteState = SS_DISCONNECTED;

					if (m_localState == SS_DISCONNECTED && m_remoteState == SS_DISCONNECTED)
					{
						m_timer.cancel();
						m_closed = true;
		
						AutoUnlock lock(m_cs);
						m_proxy.deleteSession(m_id);
					}

					return;
				}

				if (m_sendBufLocal.empty())
					return;
			}


			DbgPrint("NFSession::send_local_complete[%llu] send %d", m_id, m_sendBufLocal.front()->size());

			if (!m_sendLocalInProgress)
			{
				m_sendLocalInProgress = true;
				
				boost::asio::async_write(m_socketLocal,
						  boost::asio::buffer(&(*m_sendBufLocal.front())[0], m_sendBufLocal.front()->size()),
						  boost::bind(&NFSession::send_local_complete, 
							shared_from_this(),
							boost::asio::placeholders::error,
							boost::asio::placeholders::bytes_transferred));
			}
		} else
		{
			DbgPrint("NFSession::send_local_complete[%llu] aborted", m_id);

			m_remoteState = SS_DISCONNECTED;
			m_sendLocalInProgress = false;

			while (!m_sendBufLocal.empty())
				m_sendBufLocal.pop();

			{
				AutoUnlock lock(m_cs);
				if (m_eventsDisabled)
				{
					_postSend(NULL, 0);
				} else
				{
					if (m_ci.direction == NF_D_OUT)
						m_pEventHandler->tcpSend(m_id, NULL, 0);
					else
						m_pEventHandler->tcpReceive(m_id, NULL, 0);
				}
			}

			setStateInternal(m_wantSuspend);

			if (m_localState == SS_DISCONNECTED && m_remoteState == SS_DISCONNECTED)
			{
				m_timer.cancel();
				m_closed = true;

				AutoUnlock lock(m_cs);
				m_proxy.deleteSession(m_id);
			}

		}
	}

	void send_remote_complete(const boost::system::error_code& error, size_t bytes_transferred)
	{
		AutoLock lock(m_cs);

		if (!error)
		{
			DbgPrint("NFSession::send_remote_complete[%llu]", m_id);

			m_sendRemoteInProgress = false;

			setTimeout();

			if (!m_sendBufRemote.empty())
				m_sendBufRemote.pop();

			if (((m_sendBufLocal.size() > 4) || (m_sendBufRemote.size() > 4)) &&
				(m_localState == SS_NONE && m_remoteState == SS_NONE))
			{
			} else
			{
				setStateInternal(m_wantSuspend);
			}

			if (m_sendBufRemote.empty())
			{
				if (!m_eventsDisabled)
				{
					AutoUnlock lock(m_cs);

					if (m_ci.direction == NF_D_OUT)
						m_pEventHandler->tcpCanSend(m_id);
					else
						m_pEventHandler->tcpCanReceive(m_id);

				}

				if (m_sendBufRemote.empty() && (m_localState == SS_MUST_CLOSE))
				{
					DbgPrint("NFSession::send_remote_complete[%llu] disconnecting", m_id);

					try {
						m_socketRemote.shutdown(boost::asio::socket_base::shutdown_send);
					} catch (std::exception &)
					{
					}

					m_localState = SS_DISCONNECTED;

					if (m_localState == SS_DISCONNECTED && m_remoteState == SS_DISCONNECTED)
					{
						m_timer.cancel();
						m_closed = true;
		
						AutoUnlock lock(m_cs);
						m_proxy.deleteSession(m_id);
					}

					return;
				}

				if (m_sendBufRemote.empty())
					return;
			}
		
			if (!m_sendRemoteInProgress)
			{
				m_sendRemoteInProgress = true;

				boost::asio::async_write(m_socketRemote,
						  boost::asio::buffer(&(*m_sendBufRemote.front())[0], m_sendBufRemote.front()->size()),
						  boost::bind(&NFSession::send_remote_complete, 
							shared_from_this(),
							boost::asio::placeholders::error,
							boost::asio::placeholders::bytes_transferred));
			}
		} else
		{
			DbgPrint("NFSession::send_remote_complete[%llu] disconnected, err=%d", m_id, error.value());

			m_localState = SS_DISCONNECTED;
			m_sendRemoteInProgress = false;

			while (!m_sendBufRemote.empty())
				m_sendBufRemote.pop();

			{
				AutoUnlock lock(m_cs);
	
				if (m_eventsDisabled)
				{
					_postReceive(NULL, 0);
				} else
				{
					if (m_ci.direction == NF_D_OUT)
						m_pEventHandler->tcpReceive(m_id, NULL, 0);
					else
						m_pEventHandler->tcpSend(m_id, NULL, 0);
				}
			}

			setStateInternal(m_wantSuspend);

			if (m_localState == SS_DISCONNECTED && m_remoteState == SS_DISCONNECTED)
			{
				m_timer.cancel();
				m_closed = true;

				AutoUnlock lock(m_cs);
				m_proxy.deleteSession(m_id);
			}
		}
	}

	bool setSocketOptions(int s)
	{
		int optVal = 1;

	 	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &optVal, sizeof(optVal)) == -1)
		{
			DbgPrint("NFSession::setSocketOptions socket %d, SO_REUSEADDR error %d", s, errno);
        	return false;
	    }

		optVal = MARK_OUT;
	 	if (setsockopt(s, SOL_SOCKET, SO_MARK, &optVal, sizeof(optVal)) == -1)
		{
			DbgPrint("NFSession::setSocketOptions SO_MARK error %d", errno);
        	return false;
	    }

		return true;
	}

    bool isLocalAddress(const sockaddr * pAddr)
	{
		struct ifaddrs *ifaddr, *ifa;
		bool result = false;
		static unsigned short mapToIPv6[] = { 0, 0, 0, 0, 0, 0xffff };
		sockaddr_in mappedAddr;

		if (pAddr->sa_family == AF_INET6)
		{
			if (memcmp(&((sockaddr_in6*)pAddr)->sin6_addr, mapToIPv6, sizeof(mapToIPv6)) == 0)
			{
				memset(&mappedAddr, 0, sizeof(mappedAddr));
				mappedAddr.sin_family = AF_INET;
				memcpy(&mappedAddr.sin_addr, (char*)&((sockaddr_in6*)pAddr)->sin6_addr + 6*sizeof(short), 2*sizeof(short));
				pAddr = (sockaddr*)&mappedAddr;

				DbgPrint("NFSession::isLocalAddress mapped IPv4 %x", mappedAddr.sin_addr.s_addr);
			}
		}

		if (getifaddrs(&ifaddr) == -1) 
		{
			return false;
		}

		for (ifa = ifaddr; ifa != NULL; ifa = ifa->ifa_next) 
		{
			if ((ifa->ifa_addr == NULL) || 
				(ifa->ifa_addr->sa_family != pAddr->sa_family))
				continue;

			if (pAddr->sa_family == AF_INET)
			{
				if (memcmp(&((sockaddr_in*)pAddr)->sin_addr, &((sockaddr_in*)ifa->ifa_addr)->sin_addr, 4) == 0)
				{ 
					result = true;
					break;
				}
			} else
			{
				if (memcmp(&((sockaddr_in6*)pAddr)->sin6_addr, &((sockaddr_in6*)ifa->ifa_addr)->sin6_addr, 16) == 0)
				{ 
					result = true;
					break;
				}
			}
		}

		freeifaddrs(ifaddr);
		
		return result;
	}

	typedef std::vector<char> tBuffer;
	typedef boost::shared_ptr<tBuffer> tBuffer_ptr;
	typedef std::queue<tBuffer_ptr> tBuffers;

	enum SOCKET_STATE
	{
		SS_NONE,
		SS_DISCONNECTING,
		SS_MUST_CLOSE,
		SS_DISCONNECTED
	};

	NF_EventHandler * m_pEventHandler;
	ENDPOINT_ID	m_id;
	NF_TCP_CONN_INFO m_ci;
	unsigned int m_uid;
	NFSessionContainer &	m_proxy;
	tcp::socket m_socketLocal;
	tcp::socket m_socketRemote;
	char	m_recvBufLocal[PACKET_LEN];
	tBuffers m_sendBufLocal;
	char	m_recvBufRemote[PACKET_LEN];
	tBuffers m_sendBufRemote;
	bool	m_recvRemoteInProgress;
	bool	m_recvLocalInProgress;
	bool	m_sendRemoteInProgress;
	bool	m_sendLocalInProgress;
	bool	m_suspend;
	bool	m_wantSuspend;
	SOCKET_STATE	m_localState;
	SOCKET_STATE	m_remoteState;
	bool	m_firstPacket;
	bool	m_closed;
	boost::asio::deadline_timer m_timer;
	unsigned int m_timeout;
	bool	m_eventsDisabled;
	AutoCriticalSection m_cs;
};

typedef boost::shared_ptr<NFSession> NFSession_ptr;

class NFProxy : public boost::enable_shared_from_this<NFProxy>, public NFSessionContainer
{
public:

	NFProxy(NF_EventHandler * pEventHandler) : 
		m_pEventHandler(pEventHandler),
		m_timeout(20 * 60)
	{
		m_closing = false;
		m_sessionId = 1;

        m_io_context = std::make_unique<boost::asio::io_context>();
        m_timer = std::make_unique<boost::asio::deadline_timer>(*m_io_context);

		m_pAcceptorIPv4 = NULL;
		for (m_portIPv4 = 8080; m_portIPv4 < 8180; m_portIPv4++)
		{
			try
			{
				m_pAcceptorIPv4 = new tcp::acceptor(*m_io_context, tcp::endpoint(tcp::v4(), m_portIPv4));
				DbgPrint("NFProxy() created IPv4 acceptor");
			}
			catch (std::exception& e)
			{
				DbgPrint("NFProxy() Exception: %s", e.what());
				continue;
			}

			setSocketOptions(m_pAcceptorIPv4->native_handle());

			start_acceptIPv4();
			break;
		}

		if (m_portIPv4 == 8180)
		{
			DbgPrint("NFProxy() failed to start IPv4 proxy");
			return;
		}

		m_portIPv6 = m_portIPv4;

		m_pAcceptorIPv6 = NULL;

#ifndef NO_IPV6_NAT
		for (m_portIPv6 = 8080; m_portIPv6 < 8180; m_portIPv6++)
		{
			try
			{
				m_pAcceptorIPv6 = new tcp::acceptor(*m_io_context, tcp::endpoint(tcp::v6(), m_portIPv6));
				DbgPrint("NFProxy() created IPv6 acceptor");
			}
			catch (std::exception& e)
			{
				DbgPrint("NFProxy() Exception: %s, port=%d", e.what(), m_portIPv6);
				continue;
			}

			setSocketOptions(m_pAcceptorIPv6->native_handle());

			start_acceptIPv6();
			break;
		}

		if (m_portIPv6 == 8180)
		{
			DbgPrint("NFProxy() failed to start IPv6 proxy");
			return;
		}
#endif
	}

	~NFProxy()
	{
		if (m_pAcceptorIPv4)
			delete m_pAcceptorIPv4;
#ifndef NO_IPV6_NAT
		if (m_pAcceptorIPv6)
			delete m_pAcceptorIPv6;
#endif

         m_timer.reset();
         m_io_context.reset();
	}
  
	void start(int nThreads)
	{
		m_closing = false;

		for (std::size_t i = 0; i < nThreads; ++i)
		{
			boost::shared_ptr<boost::thread> thread(new boost::thread(
				boost::bind(&boost::asio::io_context::run, &*m_io_context)));
			
			DbgPrint("NFProxy::start() thread %d", i);

			m_threads.push_back(thread);
		}
	}

	void stop()
	{	
		closeSessions();

		m_io_context->stop();

		for (std::size_t i = 0; i < m_threads.size(); ++i)
			m_threads[i]->join();

		m_threads.clear();
	}

	void closeSessions()
	{
		m_timer->expires_from_now(boost::posix_time::seconds(0));

		m_timer->async_wait(boost::bind(&NFProxy::closeByTimer, 
			shared_from_this(),
			boost::asio::placeholders::error));

		while (!m_sessions.empty())
		{
			DbgPrint("NFProxy::closeSessions() nSessions: %d", m_sessions.size());
			sleep(1);
		}
	}

	unsigned short getPort()
	{
		return m_portIPv4;
	}

	unsigned short getPortIPv6()
	{
		return m_portIPv6;
	}

	void setTimeout(unsigned int timeout)
	{
		AutoLock lock(m_cs);
		m_timeout = timeout;
	}

	bool deleteSession_unsafe(ENDPOINT_ID id)
	{
		tSessions::iterator it = m_sessions.find(id);
		if (it == m_sessions.end())
			return false;

		m_sessionsToDelete.push(id);

		m_timer->expires_from_now(boost::posix_time::seconds(0));

		m_timer->async_wait(boost::bind(&NFProxy::deleteSessionsByTimer, 
			shared_from_this(),
			boost::asio::placeholders::error));

		return true;
	}

	bool deleteSession(ENDPOINT_ID id)
	{
		AutoLock lock(m_cs);

		return deleteSession_unsafe(id);
	}

	bool postSend(ENDPOINT_ID id, const char * buf, int len)
	{
		NFSession_ptr p;
		if (!findSession(id, p))
			return false;

		return p->postSend(buf, len);
	}

	bool postReceive(ENDPOINT_ID id, const char * buf, int len)
	{
		NFSession_ptr p;
		if (!findSession(id, p))
			return false;

		return p->postReceive(buf, len);
	}

	bool setState(ENDPOINT_ID id, bool suspend)
	{
		NFSession_ptr p;
		if (!findSession(id, p))
			return false;

		p->setState(suspend);
		return true;
	}

	bool disableEvents(ENDPOINT_ID id, bool disable)
	{
		NFSession_ptr p;
		if (!findSession(id, p))
			return false;

		p->disableEvents(disable);
		return true;
	}

	unsigned long getConnCount()
	{
		AutoLock lock(m_cs);

		return m_sessions.size();
	}

	bool getUid(ENDPOINT_ID id, unsigned int * pUid)
	{
		NFSession_ptr p;
		if (!findSession(id, p))
			return false;

		*pUid = p->getUid();
		return true;
	}

	bool getConnInfo(ENDPOINT_ID id, NF_TCP_CONN_INFO * pConnInfo)
	{
		NFSession_ptr p;
		if (!findSession(id, p))
			return false;

		*pConnInfo = p->getConnInfo();
		return true;
	}

protected:
	bool findSession(ENDPOINT_ID id, NFSession_ptr & ptr)
	{
		AutoLock lock(m_cs);
		tSessions::iterator it = m_sessions.find(id);
		if (it == m_sessions.end())
			return false;
		ptr = it->second;			
		return true;
	}

	void start_acceptIPv4()
	{
		NFSession_ptr new_session(new NFSession(*this, m_pEventHandler, *m_io_context, m_sessionId, m_timeout));
	
		{
			AutoLock lock(m_cs);
			m_sessions[m_sessionId] = new_session;
			m_sessionId++;
		}

		m_pAcceptorIPv4->async_accept(new_session->getLocalSocket(),
			boost::bind(&NFProxy::handle_acceptIPv4, this, new_session,
				boost::asio::placeholders::error));
	}

	void handle_acceptIPv4(NFSession_ptr new_session,
		const boost::system::error_code& error)
	{
		if (!error)
		{
			DbgPrint("NFProxy::handle_acceptIPv4");

			try {
				if (!new_session->start())
				{
					deleteSession(new_session->getId());
				}
			}
			catch (std::exception& e)
			{
				DbgPrint("NFProxy::handle_acceptIPv4 Exception: %s", e.what());
				deleteSession(new_session->getId());
			}

		} else
		{
			AutoLock lock(m_cs);
			m_sessions.erase(new_session->getId());
			DbgPrint("NFProxy::handle_acceptIPv4 aborted, err=%d", error.value());
		}
 		if (!m_closing)
		    start_acceptIPv4();
	}

	void start_acceptIPv6()
	{
		NFSession_ptr new_session(new NFSession(*this, m_pEventHandler, *m_io_context, m_sessionId, m_timeout));
	
		{
			AutoLock lock(m_cs);
			m_sessions[m_sessionId] = new_session;
			m_sessionId++;
		}

		m_pAcceptorIPv6->async_accept(new_session->getLocalSocket(),
			boost::bind(&NFProxy::handle_acceptIPv6, this, new_session,
				boost::asio::placeholders::error));
	}

	void handle_acceptIPv6(NFSession_ptr new_session,
		const boost::system::error_code& error)
	{
		if (!error)
		{
			DbgPrint("NFProxy::handle_acceptIPv6");

			try {
				if (!new_session->start())
				{
					deleteSession(new_session->getId());
				}
			}
			catch (std::exception& e)
			{
				DbgPrint("NFProxy::handle_acceptIPv6 Exception: %s", e.what());
				deleteSession(new_session->getId());
			}

		} else
		{
			AutoLock lock(m_cs);
			m_sessions.erase(new_session->getId());
			DbgPrint("NFProxy::handle_acceptIPv6 aborted, err=%d", error.value());
		}

 		if (!m_closing)
		    start_acceptIPv6();
	}

	void deleteSessionsByTimer(const boost::system::error_code& error)
	{
		ENDPOINT_ID id;
		tSessions::iterator it;
		NF_TCP_CONN_INFO ci;
		AutoLock lock(m_cs);
		
		while (!m_sessionsToDelete.empty())
		{
			id = m_sessionsToDelete.front();
			m_sessionsToDelete.pop();

			it = m_sessions.find(id);
			if (it == m_sessions.end())
				continue;

			ci = it->second->getConnInfo();

			DbgPrint("NFProxy::deleteSessionsByTimer[%llu]", id);

			{
				AutoUnlock lock(m_cs);
                m_pEventHandler->tcpClosed(id, &ci);
			}

			it = m_sessions.find(id);
			if (it == m_sessions.end())
				continue;

			boost::system::error_code ec;
			it->second->close(ec);

			m_sessions.erase(it);
		}
	}

	void closeByTimer(const boost::system::error_code& error)
	{
		{
			AutoLock lock(m_cs);
	
			m_closing = true;
	
			if (m_pAcceptorIPv4)
			{
				m_pAcceptorIPv4->cancel();
				m_pAcceptorIPv4->close();
			}

			if (m_pAcceptorIPv6)
			{
				m_pAcceptorIPv6->cancel();
				m_pAcceptorIPv6->close();
			}
		}

		{
			AutoLock lock(m_cs);
			boost::system::error_code ec;
			for (tSessions::iterator it = m_sessions.begin(); it != m_sessions.end(); it++)
			{
				it->second->close(ec);
				deleteSession(it->first);
			}
		}
	
	}

	bool setSocketOptions(int s)
	{
		int optVal = 1;

	 	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &optVal, sizeof(optVal)) == -1)
		{
			DbgPrint("NFProxy::setSocketOptions SO_REUSEADDR error %d", errno);
        	return false;
	    }

	    if (setsockopt(s, SOL_IP, IP_TRANSPARENT, &optVal, sizeof(optVal)) == -1)
		{
			DbgPrint("NFProxy::setSocketOptions IP_TRANSPARENT error %d", errno);
			return false;
	    }

		return true;
	}

private:
	typedef std::map<ENDPOINT_ID, NFSession_ptr> tSessions;
	tSessions m_sessions;
    std::unique_ptr<boost::asio::io_context> m_io_context;
    std::unique_ptr<boost::asio::deadline_timer> m_timer;
	tcp::acceptor * m_pAcceptorIPv4;
	tcp::acceptor * m_pAcceptorIPv6;
	ENDPOINT_ID m_sessionId;
	NF_EventHandler * m_pEventHandler;
	AutoCriticalSection m_cs;
	unsigned short m_portIPv4;
	unsigned short m_portIPv6;
	typedef std::queue<ENDPOINT_ID> tSessionsToDelete;
	tSessionsToDelete m_sessionsToDelete;
	unsigned int m_timeout;
	std::vector<boost::shared_ptr<boost::thread> > m_threads;
	bool m_closing;
};

typedef boost::shared_ptr<NFProxy> NFProxy_ptr;

#endif
