//
// 	NetFilterSDK 
// 	Copyright (C) Vitaly Sidorov
//	All rights reserved.
//
//	This file is a part of the NetFilter SDK.
//	The code and information is provided "as-is" without
//	warranty of any kind, either expressed or implied.
//

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/un.h>
#include <sys/errno.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/types.h>

#include "nfproxy.h"

extern AutoCriticalSection	g_cs;
extern NFProxy_ptr			g_proxy;

extern std::string			g_driverName;

static std::string			g_driverName_redirect_out;
static std::string			g_driverName_redirect_in;

static std::string			g_driverName_in;
static std::string			g_driverName_out;

typedef std::vector<std::string> tParamList;
typedef std::vector<tParamList> tRulesList;
typedef std::vector<const char*> tParamPtrArray;

#define FWCTRL_PROCESS_IP "ip"
#define FWCTRL_PROCESS_IPV4 "iptables"
#define FWCTRL_PROCESS_IPV6 "ip6tables"
#define FWCTRL_FILTER "filter"
#define FWCTRL_INPUT "INPUT"
#define FWCTRL_OUTPUT "OUTPUT"
#define FWCTRL_PREROUTING "PREROUTING"
#define FWCTRL_POSTROUTING "POSTROUTING"
#define FWCTRL_MANGLE "mangle"
#define FWCTRL_NAT "nat"

#if defined(_DEBUG) || defined(_RELEASE_LOG)
#define DbgPrintParamList(p) _DbgPrintParamList(p)

void _DbgPrintParamList(const tParamList & params)
{
	std::string s;

	tParamList::const_iterator it;

	for (it = params.begin(); it != params.end(); it++)
	{
		s += *it + " ";
	}

	DbgPrint("%s", s.c_str());
}

#else
#define DbgPrintParamList(p)
#endif

static int do_exec(char *const args[])
{
	pid_t pid;
	int status;
	int eo;

	pid = fork();
	if (pid == -1) 
	{
		DbgPrint("fork failed: %s", strerror(errno));
		return -1;
	}
	if (!pid) 
	{
		int fd = open("/dev/null", O_WRONLY);
        if (fd > 0)
		{
			if (dup2(fd, 2)  != 1 )
			{
               //handle error 
			}
        }

		/* In child process, execute external command. */
		(void)execvp(args[0], args);
		/* We only get here if the execv() failed. */
		eo = errno;
//		DbgPrint("execvp %s failed: %s", args[0], strerror(eo));
		_exit(eo);
	}
	/* In parent process, wait for exernal command to finish. */
	if (wait4(pid, (int*)&status, 0, NULL) != pid) 
	{
		DbgPrint("BUG executing %s command.", args[0]);
		return -1;
	}
	if (!WIFEXITED(status)) 
	{
		DbgPrint("%s command aborted by signal %d.",
				args[0], WTERMSIG(status));
		return -1;
	}
	eo = WEXITSTATUS(status);
	if (eo) 
	{
		DbgPrint("%s command failed: %s", 
				args[0], strerror(eo));
		return -1;
	}
	return 0;
}

static bool execParams(const tParamList & params)
{
	tParamPtrArray ptrArray;
	tParamList::const_iterator it;

	DbgPrint("execParams()");
	DbgPrintParamList(params);

	for (it = params.begin(); it != params.end(); it++)
	{
		ptrArray.push_back(it->c_str());
	}

	ptrArray.push_back(NULL);

	return do_exec((char*const*)&ptrArray[0]) == 0;
}

static tParamList createParamList(const char * processName, ...)
{
	tParamList params;
	va_list args;
	char * p;
	
	params.push_back(processName);

	// Add -w parameter to make iptables wait for exclusive lock
	params.push_back("-w");

	va_start(args, processName);

	p = va_arg(args, char*);
	while (p != NULL)
	{
		params.push_back(p);
		p = va_arg(args, char*);
	}

	va_end(args);

	return params;
}

static tParamList createParamListGeneric(const char * processName, ...)
{
	tParamList params;
	va_list args;
	char * p;
	
	params.push_back(processName);

	va_start(args, processName);

	p = va_arg(args, char*);
	while (p != NULL)
	{
		params.push_back(p);
		p = va_arg(args, char*);
	}

	va_end(args);

	return params;
}

static std::string subNetToString(int ipFamily, const char * ipAddr, unsigned char ipAddrMask)
{
	char buf[100];
	std::string sAddr;

	if (ipFamily == AF_INET)
	{
		if (inet_ntop(AF_INET, ipAddr, buf, sizeof(buf)))
		{
			sAddr = buf;
		}
	} else
	{
		if (inet_ntop(AF_INET6, ipAddr, buf, sizeof(buf)))
		{
			sAddr = buf;
		}
	}

	if (ipAddrMask != 0)
	{
		sAddr += "/";
		_snprintf(buf, sizeof(buf), "%d", ipAddrMask);
		sAddr += buf;
	}

	return sAddr;
}

static std::string portRangeToString(PNFEXT_PORT_RANGE pPorts)
{
	char buf[100];

	if (pPorts->from == pPorts->to)
	{
		_snprintf(buf, sizeof(buf), "%d", pPorts->from);
		return buf;
	} else
	{
		_snprintf(buf, sizeof(buf), "%d:%d", pPorts->from, pPorts->to);
		return buf;
	}
}

static std::string intToString(int n)
{
	char buf[100];
	_snprintf(buf, sizeof(buf), "%d", n);
	return buf;
}

static bool createRuleParamList(PNFEXT_RULE pRule,
								int ip_family,
								const char * ctrlProcess,
								const char * table,
								const char * chain,
								tParamList & params)
{
	DbgPrint("createRuleParamLists()");

	params = createParamList(
					ctrlProcess,
					"-t", table,
					"-A", chain,
					(char*)NULL);

	switch (pRule->protocol)
	{
	case 0:
		break;
	case IPPROTO_TCP:
		{
			params.push_back("-p");
			params.push_back("tcp");
		}
		break;
	case IPPROTO_UDP:
		{
			params.push_back("-p");
			params.push_back("udp");
		}
		break;
	default:
		{
			params.push_back("-p");
			params.push_back(intToString(pRule->protocol));
		}
		break;
	}

	switch (pRule->filteringFlag)
	{
	case NFEXT_BYPASS:
		{
			if (pRule->ruleType == NFEXT_REDIRECT_RULE)
			{
				params.push_back("-j");
				params.push_back("RETURN");
			} else
			{
				params.push_back("-j");
				params.push_back("ACCEPT");
			}
		}
		break;
	
	case NFEXT_BLOCK:
		{
			if (pRule->ruleType == NFEXT_REDIRECT_RULE)
			{
				if (pRule->protocol == 0)
				{
					params.push_back("-p");
					params.push_back("tcp");
				}
			}

			params.push_back("-j");
			params.push_back("DROP");
		}
		break;
	
	case NFEXT_REDIRECT:
		{
			if (pRule->ruleType != NFEXT_REDIRECT_RULE)
			{
				DbgPrint("createRuleParamLists() invalid rule type %d", pRule->ruleType);
				return false;
			}

			if (pRule->protocol == 0)
			{
				params.push_back("-p");
				params.push_back("tcp");
			}

			params.push_back("-m");
			params.push_back("mark");
//			params.push_back("!");
			params.push_back("--mark");
			params.push_back("0/"MARK_PASS_S);
			params.push_back("-m");
			params.push_back("conntrack");
			params.push_back("--ctstate");
			params.push_back("NEW,ESTABLISHED");
			params.push_back("--ctdir");
			params.push_back("ORIGINAL");
			params.push_back("-j");
			params.push_back("MARK");
			params.push_back("--set-xmark");
			params.push_back(MARK_IN_S"/0xffffffff");
		}
		break;

	default:
		DbgPrint("createRuleParamLists() invalid filtering flag %d", pRule->filteringFlag);
		return false;
	}

	if (pRule->fieldsMask & NFEXT_USE_DST_IP)
	{
		params.push_back("--dst");
		params.push_back(subNetToString(ip_family, pRule->dstIp, pRule->dstIpMask));
	}

	if (pRule->fieldsMask & NFEXT_USE_SRC_IP)
	{
		params.push_back("--src");
		params.push_back(subNetToString(ip_family, pRule->srcIp, pRule->srcIpMask));
	}

	if (pRule->fieldsMask & NFEXT_USE_DST_PORTS)
	{
		params.push_back("--dport");
		params.push_back(portRangeToString(&pRule->dstPorts));
	}

	if (pRule->fieldsMask & NFEXT_USE_SRC_PORTS)
	{
		params.push_back("--sport");
		params.push_back(portRangeToString(&pRule->srcPorts));
	}

	if (pRule->fieldsMask & (NFEXT_USE_UID | NFEXT_USE_GID))
	{
		params.push_back("-m");
		params.push_back("owner");
	}

	if (pRule->fieldsMask & NFEXT_USE_UID)
	{
		params.push_back("--uid-owner");
		params.push_back(intToString(pRule->uid));
	}

	if (pRule->fieldsMask & NFEXT_USE_GID)
	{
		params.push_back("--gid-owner");
		params.push_back(intToString(pRule->gid));
	}

	return true;
}


static bool createRuleParamLists(PNFEXT_RULE pRule,
								 int ip_family,
								tRulesList & list)
{
	if (pRule->ip_family != 0 &&
		pRule->ip_family != ip_family)
	{
		return true;
	}

	const char * ctrlProcess = 
		(ip_family == AF_INET)? 
		FWCTRL_PROCESS_IPV4 : FWCTRL_PROCESS_IPV6;

	if (pRule->ruleType == NFEXT_REDIRECT_RULE)
	{
#ifdef NO_IPV6_NAT
		if (ip_family == AF_INET6)
		{
			return true;
		}
#endif
		if (pRule->direction == NF_D_IN || pRule->direction == NF_D_BOTH || pRule->direction == 0)
		{
			tParamList params;
			if (!createRuleParamList(pRule, ip_family, ctrlProcess, FWCTRL_MANGLE, g_driverName_redirect_in.c_str(), params))
			{
				return false;
			}
			list.push_back(params);
		}

		if (pRule->direction == NF_D_OUT || pRule->direction == NF_D_BOTH || pRule->direction == 0)
		{
			tParamList params;
			if (!createRuleParamList(pRule, ip_family, ctrlProcess, FWCTRL_MANGLE, g_driverName_redirect_out.c_str(), params))
			{
				return false;
			}
			list.push_back(params);
		}
	} else
	if (pRule->ruleType == NFEXT_PACKET_RULE)
	{
		if (pRule->direction == NF_D_IN || pRule->direction == NF_D_BOTH || pRule->direction == 0)
		{
			tParamList params;
			if (!createRuleParamList(pRule, ip_family, ctrlProcess, FWCTRL_FILTER, g_driverName_in.c_str(), params))
			{
				return false;
			}
			list.push_back(params);
		}

		if (pRule->direction == NF_D_OUT || pRule->direction == NF_D_BOTH || pRule->direction == 0)
		{
			tParamList params;
			if (!createRuleParamList(pRule, ip_family, ctrlProcess, FWCTRL_FILTER, g_driverName_out.c_str(), params))
			{
				return false;
			}
			list.push_back(params);
		}
	} else
	{
		DbgPrint("createRuleParamLists() failed, invalid rule type %d", pRule->ruleType);
		return false;
	}

	return true;
}

static bool execRulesList(tRulesList & list, bool stopOnError = true)
{
	for (tRulesList::iterator it = list.begin(); it != list.end(); it++)
	{
		if (!execParams(*it))
		{
			DbgPrint("execRulesList() failed");
			
			if (stopOnError)
				return false;
		}
	}

	return true;
}

NFAPI_API NF_STATUS NFAPI_NS 
nf_setRules(PNFEXT_RULE pRules, int count)
{
	tRulesList ipv4RulesList;
	tRulesList ipv6RulesList;

	g_driverName_in = g_driverName + "_in";
	g_driverName_out = g_driverName + "_out";
	g_driverName_redirect_in = g_driverName + "_rd_in";
	g_driverName_redirect_out = g_driverName + "_rd_out";

	nf_deleteRules();

	AutoLock lock(g_cs);

	std::string sUid = intToString(geteuid());

	DbgPrint("nf_setRules");

	ipv4RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_MANGLE,
				"-N", g_driverName_redirect_out.c_str(),
				(char*)NULL));

	ipv4RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_MANGLE,
				"-N", g_driverName_redirect_in.c_str(),
				(char*)NULL));

#ifndef NO_IPV6_NAT

	ipv6RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_MANGLE,
				"-N", g_driverName_redirect_out.c_str(),
				(char*)NULL));

	ipv6RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_MANGLE,
				"-N", g_driverName_redirect_in.c_str(),
				(char*)NULL));

	ipv6RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_MANGLE,
				"-A", g_driverName_redirect_out.c_str(),
				"-p", "tcp",
				"-m", "owner",
				"--uid-owner", sUid.c_str(),
				"-j", "RETURN",
				(char*)NULL));

	ipv6RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_MANGLE,
				"-A", g_driverName_redirect_out.c_str(),
				"-p", "tcp",
				"--src", "::ffff:127.0.0.1",
				"-j", "RETURN",
				(char*)NULL));

	ipv6RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_MANGLE,
				"-A", g_driverName_redirect_in.c_str(),
				"-p", "tcp",
				"--src", "::ffff:127.0.0.1",
				"-j", "RETURN",
				(char*)NULL));
#endif
	
	// Bypass Docker traffic

	ipv4RulesList.push_back(
		createParamList(
				ctrlProcessIPv4,
				"-t", FWCTRL_MANGLE,
				"-A", g_driverName_redirect_in.c_str(),
				"-i", "br-+",
				"-j", "RETURN",
				(char*)NULL));

	ipv6RulesList.push_back(
		createParamList(
				ctrlProcessIPv6,
				"-t", FWCTRL_MANGLE,
				"-A", g_driverName_redirect_in.c_str(),
				"-i", "br-+",
				"-j", "RETURN",
				(char*)NULL));

	ipv4RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_MANGLE,
				"-A", g_driverName_redirect_in.c_str(),
				"-i", "docker0",
				"-j", "RETURN",
				(char*)NULL));

	ipv6RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_MANGLE,
				"-A", g_driverName_redirect_in.c_str(),
				"-i", "docker0",
				"-j", "RETURN",
				(char*)NULL));

	// Bypass LXC traffic

	ipv4RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_MANGLE,
				"-A", g_driverName_redirect_in.c_str(),
				"-i", "lxcbr0",
				"-j", "RETURN",
				(char*)NULL));

	ipv6RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_MANGLE,
				"-A", g_driverName_redirect_in.c_str(),
				"-i", "lxcbr0",
				"-j", "RETURN",
				(char*)NULL));

	ipv4RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_FILTER,
				"-N", g_driverName_in.c_str(),
				(char*)NULL));

	ipv4RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_FILTER,
				"-N", g_driverName_out.c_str(),
				(char*)NULL));

	ipv4RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_MANGLE,
				"-A", g_driverName_redirect_out.c_str(),
				"-p", "tcp",
				"-m", "owner",
				"--uid-owner", sUid.c_str(),
				"-j", "RETURN",
				(char*)NULL));

	ipv4RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_MANGLE,
				"-A", g_driverName_redirect_out.c_str(),
				"-p", "tcp",
				"--src", "127.0.0.1",
				"-j", "RETURN",
				(char*)NULL));

	ipv4RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_MANGLE,
				"-A", g_driverName_redirect_in.c_str(),
				"-p", "tcp",
				"--src", "127.0.0.1",
				"-j", "RETURN",
				(char*)NULL));

	ipv6RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_FILTER,
				"-N", g_driverName_in.c_str(),
				(char*)NULL));

	ipv6RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_FILTER,
				"-N", g_driverName_out.c_str(),
				(char*)NULL));

	{
		NFEXT_RULE rule;
	
		memset(&rule, 0, sizeof(NFEXT_RULE));
		rule.ruleType = NFEXT_PACKET_RULE;
		rule.protocol = IPPROTO_TCP;
		rule.direction = NF_D_IN;
		rule.fieldsMask = NFEXT_USE_DST_PORTS;
		rule.dstPorts.from = rule.dstPorts.to = g_proxy->getPort();
		rule.filteringFlag = NFEXT_BYPASS;

		if (!createRuleParamLists(&rule, AF_INET, ipv4RulesList))
		{
			DbgPrint("nf_setRules() unable to add allowing rule for proxy");
			return NF_STATUS_FAIL;
		}
	}

	{
		NFEXT_RULE rule;
	
		memset(&rule, 0, sizeof(NFEXT_RULE));
		rule.ruleType = NFEXT_PACKET_RULE;
		rule.protocol = IPPROTO_TCP;
		rule.direction = NF_D_IN;
		rule.fieldsMask = NFEXT_USE_DST_PORTS;
		rule.dstPorts.from = rule.dstPorts.to = g_proxy->getPortIPv6();
		rule.filteringFlag = NFEXT_BYPASS;

		if (!createRuleParamLists(&rule, AF_INET6, ipv6RulesList))
		{
			DbgPrint("nf_setRules() unable to add allowing rule for proxy (IPv6)");
			return NF_STATUS_FAIL;
		}
	}

	for (int i=0; i<count; i++)
	{
		if (!createRuleParamLists(pRules+i, AF_INET, ipv4RulesList))
		{
			DbgPrint("nf_setRules() invalid IPv4 rule %d", i);
			return NF_STATUS_FAIL;
		}
		if (!createRuleParamLists(pRules+i, AF_INET6, ipv6RulesList))
		{
			DbgPrint("nf_setRules() invalid IPv6 rule %d", i);
			return NF_STATUS_FAIL;
		}
	}

	ipv4RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_MANGLE,
				"-A", g_driverName_redirect_in.c_str(),
				"-p", "tcp",
				"-m", "mark",
				"--mark", MARK_IN_S"/"MARK_IN_S,
				"-j", "TPROXY",
				"--on-port", intToString(g_proxy->getPort()).c_str(),
				"--on-ip", "127.0.0.1",
				"--tproxy-mark", "0x0/0x0",
				(char*)NULL));

	ipv6RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_MANGLE,
				"-A", g_driverName_redirect_in.c_str(),
				"-p", "tcp",
				"-m", "mark",
				"--mark", MARK_IN_S"/"MARK_IN_S,
				"-j", "TPROXY",
				"--on-port", intToString(g_proxy->getPortIPv6()).c_str(),
				"--on-ip", "::ffff:127.0.0.1",
				"--tproxy-mark", "0x0/0x0",
				(char*)NULL));

	ipv4RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_MANGLE,
				"-A", FWCTRL_OUTPUT,
				"-j", g_driverName_redirect_out.c_str(),
				(char*)NULL));

	ipv4RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_MANGLE,
				"-A", FWCTRL_PREROUTING,
				"-j", g_driverName_redirect_in.c_str(),
				(char*)NULL));

#ifndef NO_IPV6_NAT

	ipv6RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_MANGLE,
				"-A", FWCTRL_OUTPUT,
				"-j", g_driverName_redirect_out.c_str(),
				(char*)NULL));

	ipv6RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_MANGLE,
				"-A", FWCTRL_PREROUTING,
				"-j", g_driverName_redirect_in.c_str(),
				(char*)NULL));

#endif

	ipv4RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_FILTER,
				"-I", FWCTRL_INPUT,
				"-j", g_driverName_in.c_str(),
				(char*)NULL));

	ipv4RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_FILTER,
				"-I", FWCTRL_OUTPUT,
				"-j", g_driverName_out.c_str(),
				(char*)NULL));

	ipv6RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_FILTER,
				"-I", FWCTRL_INPUT,
				"-j", g_driverName_in.c_str(),
				(char*)NULL));

	ipv6RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_FILTER,
				"-I", FWCTRL_OUTPUT,
				"-j", g_driverName_out.c_str(),
				(char*)NULL));

	ipv4RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_NAT,
				"-I", FWCTRL_OUTPUT,
				"-p", "tcp",
				"-m", "mark",
				"--mark", MARK_IN_S"/"MARK_IN_S,
				"-j", "ACCEPT",
				(char*)NULL));

	ipv4RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_NAT,
				"-I", FWCTRL_POSTROUTING,
				"-p", "tcp",
				"-m", "mark",
				"--mark", MARK_IN_S"/"MARK_IN_S,
				"-j", "ACCEPT",
				(char*)NULL));

#ifndef NO_IPV6_NAT

	ipv6RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_NAT,
				"-I", FWCTRL_OUTPUT,
				"-p", "tcp",
				"-m", "mark",
				"--mark", MARK_IN_S"/"MARK_IN_S,
				"-j", "ACCEPT",
				(char*)NULL));

	ipv6RulesList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_NAT,
				"-I", FWCTRL_POSTROUTING,
				"-p", "tcp",
				"-m", "mark",
				"--mark", MARK_IN_S"/"MARK_IN_S,
				"-j", "ACCEPT",
				(char*)NULL));
#endif

	if (!execRulesList(ipv4RulesList))
	{
		DbgPrint("nf_setRules() unable to add IPv4 rules");
		return NF_STATUS_FAIL;
	}

	if (!execRulesList(ipv6RulesList))
	{
		DbgPrint("nf_setRules() unable to add IPv6 rules");
//		return NF_STATUS_FAIL;
	}




	return NF_STATUS_SUCCESS;
}

NFAPI_API NF_STATUS NFAPI_NS 
nf_deleteRules()
{
	AutoLock lock(g_cs);
	
	tRulesList ipv4RulesDeleteList;
	tRulesList ipv6RulesDeleteList;

	ipv4RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_MANGLE,
				"-F", g_driverName_redirect_out.c_str(),
				(char*)NULL));

	ipv4RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_MANGLE,
				"-F", g_driverName_redirect_in.c_str(),
				(char*)NULL));

	ipv4RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_MANGLE,
				"-D", FWCTRL_OUTPUT,
				"-j", g_driverName_redirect_out.c_str(),
				(char*)NULL));

	ipv4RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_MANGLE,
				"-D", FWCTRL_PREROUTING,
				"-j", g_driverName_redirect_in.c_str(),
				(char*)NULL));

	ipv4RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_MANGLE,
				"-X", g_driverName_redirect_out.c_str(),
				(char*)NULL));

	ipv4RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_MANGLE,
				"-X", g_driverName_redirect_in.c_str(),
				(char*)NULL));

#ifndef NO_IPV6_NAT

	ipv6RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_MANGLE,
				"-F", g_driverName_redirect_out.c_str(),
				(char*)NULL));

	ipv6RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_MANGLE,
				"-F", g_driverName_redirect_in.c_str(),
				(char*)NULL));

	ipv6RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_MANGLE,
				"-D", FWCTRL_OUTPUT,
				"-j", g_driverName_redirect_out.c_str(),
				(char*)NULL));

	ipv6RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_MANGLE,
				"-D", FWCTRL_PREROUTING,
				"-j", g_driverName_redirect_in.c_str(),
				(char*)NULL));

	ipv6RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_MANGLE,
				"-X", g_driverName_redirect_out.c_str(),
				(char*)NULL));

	ipv6RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_MANGLE,
				"-X", g_driverName_redirect_in.c_str(),
				(char*)NULL));
#endif

	ipv4RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_FILTER,
				"-F", g_driverName_in.c_str(),
				(char*)NULL));

	ipv4RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_FILTER,
				"-D", FWCTRL_INPUT,
				"-j", g_driverName_in.c_str(),
				(char*)NULL));

	ipv4RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_FILTER,
				"-X", g_driverName_in.c_str(),
				(char*)NULL));

	ipv4RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_FILTER,
				"-F", g_driverName_out.c_str(),
				(char*)NULL));

	ipv4RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_FILTER,
				"-D", FWCTRL_OUTPUT,
				"-j", g_driverName_out.c_str(),
				(char*)NULL));

	ipv4RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_FILTER,
				"-X", g_driverName_out.c_str(),
				(char*)NULL));

	ipv6RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_FILTER,
				"-F", g_driverName_in.c_str(),
				(char*)NULL));

	ipv6RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_FILTER,
				"-D", FWCTRL_INPUT,
				"-j", g_driverName_in.c_str(),
				(char*)NULL));

	ipv6RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_FILTER,
				"-X", g_driverName_in.c_str(),
				(char*)NULL));

	ipv6RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_FILTER,
				"-F", g_driverName_out.c_str(),
				(char*)NULL));

	ipv6RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_FILTER,
				"-D", FWCTRL_OUTPUT,
				"-j", g_driverName_out.c_str(),
				(char*)NULL));

	ipv6RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_FILTER,
				"-X", g_driverName_out.c_str(),
				(char*)NULL));

	ipv4RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_NAT,
				"-D", FWCTRL_OUTPUT,
				"-p", "tcp",
				"-m", "mark",
				"--mark", MARK_IN_S"/"MARK_IN_S,
				"-j", "ACCEPT",
				(char*)NULL));

	ipv4RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV4,
				"-t", FWCTRL_NAT,
				"-D", FWCTRL_POSTROUTING,
				"-p", "tcp",
				"-m", "mark",
				"--mark", MARK_IN_S"/"MARK_IN_S,
				"-j", "ACCEPT",
				(char*)NULL));

#ifndef NO_IPV6_NAT

	ipv6RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_NAT,
				"-D", FWCTRL_OUTPUT,
				"-p", "tcp",
				"-m", "mark",
				"--mark", MARK_IN_S"/"MARK_IN_S,
				"-j", "ACCEPT",
				(char*)NULL));

	ipv6RulesDeleteList.push_back(
		createParamList(
				FWCTRL_PROCESS_IPV6,
				"-t", FWCTRL_NAT,
				"-D", FWCTRL_POSTROUTING,
				"-p", "tcp",
				"-m", "mark",
				"--mark", MARK_IN_S"/"MARK_IN_S,
				"-j", "ACCEPT",
				(char*)NULL));
#endif

	if (!execRulesList(ipv4RulesDeleteList, false))
	{
		DbgPrint("nf_deleteRules() unable to delete IPv4 rules");
		return NF_STATUS_FAIL;
	}

	if (!execRulesList(ipv6RulesDeleteList, false))
	{
		DbgPrint("nf_deleteRules() unable to delete IPv6 rules");
//		return NF_STATUS_FAIL;
	}

	return NF_STATUS_SUCCESS;
}


NF_STATUS nf_addRouteRules()
{
	AutoLock lock(g_cs);
	
	tRulesList ipv4RulesList;
	tRulesList ipv6RulesList;

	ipv4RulesList.push_back(
		createParamListGeneric(
				FWCTRL_PROCESS_IP,
				"route", "add",
				"local", "default",
				"dev", "lo",
				"table", MARK_TABLE,
				(char*)NULL));

	ipv4RulesList.push_back(
		createParamListGeneric(
				FWCTRL_PROCESS_IP,
				"rule", "add",
				"fwmark", MARK_IN_S"/"MARK_IN_S,
				"lookup", MARK_TABLE,
				(char*)NULL));

	ipv6RulesList.push_back(
		createParamListGeneric(
				FWCTRL_PROCESS_IP,
				"-6",
				"route", "add",
				"local", "default",
				"dev", "lo",
				"table", MARK_TABLE,
				(char*)NULL));

	ipv6RulesList.push_back(
		createParamListGeneric(
				FWCTRL_PROCESS_IP,
				"-6",
				"rule", "add",
				"fwmark", MARK_IN_S"/"MARK_IN_S,
				"lookup", MARK_TABLE,
				(char*)NULL));

	if (!execRulesList(ipv4RulesList, false))
	{
		DbgPrint("nf_addRouteRules() unable to add IPv4 rules");
		return NF_STATUS_FAIL;
	}

	if (!execRulesList(ipv6RulesList, false))
	{
		DbgPrint("nf_addRouteRules() unable to add IPv6 rules");
//		return NF_STATUS_FAIL;
	}

	return NF_STATUS_SUCCESS;
}

NF_STATUS nf_deleteRouteRules()
{
	AutoLock lock(g_cs);
	
	tRulesList ipv4RulesList;
	tRulesList ipv6RulesList;

	ipv4RulesList.push_back(
		createParamListGeneric(
				FWCTRL_PROCESS_IP,
				"route", "del",
				"local", "default",
				"dev", "lo",
				"table", "800",
				(char*)NULL));

	ipv4RulesList.push_back(
		createParamListGeneric(
				FWCTRL_PROCESS_IP,
				"rule", "del",
				"fwmark", "0x800/0x800",
				"lookup", "800",
				(char*)NULL));

	ipv6RulesList.push_back(
		createParamListGeneric(
				FWCTRL_PROCESS_IP,
				"-6",
				"route", "del",
				"local", "default",
				"dev", "lo",
				"table", "800",
				(char*)NULL));

	ipv6RulesList.push_back(
		createParamListGeneric(
				FWCTRL_PROCESS_IP,
				"-6",
				"rule", "del",
				"fwmark", "0x800/0x800",
				"lookup", "800",
				(char*)NULL));

	if (!execRulesList(ipv4RulesList, false))
	{
		DbgPrint("nf_deleteRouteRules() unable to delete IPv4 rules");
		return NF_STATUS_FAIL;
	}

	if (!execRulesList(ipv6RulesList, false))
	{
		DbgPrint("nf_deleteRouteRules() unable to delete IPv6 rules");
//		return NF_STATUS_FAIL;
	}

	return NF_STATUS_SUCCESS;
}
